from abc import ABC, abstractmethod
from typing import Optional, List, Dict, Tuple, Any, Union
import onnxruntime
import numpy as np

from pipez.core.batch import Batch
from pipez.core.node import Node
from pipez.utils.resize import resize


class BaseORT(Node, ABC):
    def __init__(
            self,
            model_path: str,
            providers: List[Union[str, Tuple[str, Dict]]],
            pad_value: int = 0,
            half_precision: bool = False,
            **kwargs
    ):
        super().__init__(**kwargs)
        self._session = onnxruntime.InferenceSession(path_or_bytes=model_path, providers=providers)
        self._pad_value = pad_value
        self._dtype = np.float16 if half_precision else np.float32

        net_input = self._session.get_inputs()[0]
        self._input_name = net_input.name
        self._batch_size = net_input.shape[0]

        if isinstance(self._batch_size, str):
            self._batch_size = 32

        self._size = (net_input.shape[-1], net_input.shape[-2])
        self._batch = np.ones((self._batch_size, net_input.shape[1], self._size[1], self._size[0]), dtype=self._dtype)

    def _preprocessing(self, image: np.ndarray) -> Tuple[np.ndarray, Dict]:
        meta = dict(original_shape=image.shape)
        image = resize(image, self._size, self._pad_value)
        meta.update(current_shape=image.shape)
        image = self.preprocessing(image)

        return image, meta

    @abstractmethod
    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        pass

    @abstractmethod
    def preprocessing(self, image: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def postprocessing(self, output: List, meta: Dict) -> Any:
        pass



class Memory(object):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
            cls._instance._memory = {}

        return cls._instance

    def __getitem__(self, item):
        if item not in self._memory:
            raise KeyError(f'Key «{item}» not found in memory')

        return self._memory[item]

    def __setitem__(self, key, value):
        self._memory[key] = value

    def __delitem__(self, key):
        del self._memory[key]

    def __contains__(self, item):
        return item in self._memory

    def __str__(self):
        return str(self._memory)

    def __del__(self):
        del self._memory



from typing import Optional
from pipez.core.batch import Batch
from pipez.core.node import Node
from pipez.core.registry import Registry


@Registry.add
class Ungroup(Node):
    def __init__(
            self,
            class_name: str,
            **kwargs
    ):
        super().__init__(name=self.__class__.__name__, **kwargs)
        self._class_name = class_name

    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        batch = Batch(meta=data.meta)
        batch.meta.update(idxs=[], batch_size=len(data))

        for idx, obj in enumerate(data):
            for crop in obj.get(self._class_name, []):
                batch.append(crop['crop'])
                batch.meta['idxs'].append(idx)

        return batch



from abc import ABC, abstractmethod
from typing import Optional, List, Dict, Any
import numpy as np

from .base import BaseORT
from pipez.core.batch import Batch


class SyncORT(BaseORT, ABC):
    def __init__(
            self,
            data_key: Optional[str] = None,
            **kwargs
    ):
        super().__init__(**kwargs)
        self._data_key = data_key

    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        images = []
        metas = []

        for obj in data:
            image, meta = self._preprocessing(obj[self._data_key] if isinstance(obj, dict) else obj)
            images.append(image)
            metas.append(meta)

        batch = Batch(meta=data.meta)

        for i in range(0, len(images), self._batch_size):
            for batch_idx, image in enumerate(images[i: i + self._batch_size]):
                self._batch[batch_idx] = image

            inference = self._session.run(None, {self._input_name: self._batch})

            for *output, meta in zip(*inference, metas[i: i + self._batch_size]):
                batch.append(self.postprocessing(output, meta))

        return batch

    @abstractmethod
    def preprocessing(self, image: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def postprocessing(self, output: List, meta: Dict) -> Any:
        pass



from enum import Enum, auto


class NodeType(Enum):
    THREAD = auto()
    PROCESS = auto()

    @staticmethod
    def from_string(type: str) -> 'NodeType':
        if type.lower() == 'thread':
            return NodeType.THREAD
        elif type.lower() == 'process':
            return NodeType.PROCESS
        else:
            raise ValueError


class NodeStatus(Enum):
    ALIVE = auto()
    FINISH = auto()
    TERMINATE = auto()


class BatchStatus(Enum):
    OK = auto()
    END = auto()
    ERROR = auto()



from typing import Optional, List
from threading import Thread
from fastapi import FastAPI, APIRouter, Request
from fastapi.templating import Jinja2Templates
from fastapi.staticfiles import StaticFiles
from fastapi.responses import HTMLResponse
from statistics import mean, pstdev
from datetime import datetime
from os.path import join, dirname, abspath
import uvicorn
import logging

from pipez.core.batch import Batch
from pipez.core.enums import BatchStatus
from pipez.core.node import Node
from pipez.core.queue_wrapper import QueueWrapper


class Watchdog(Node):
    def __init__(
            self,
            pipeline: List[Node],
            *,
            verbose_metrics: bool = False,
            metrics_host: str = '0.0.0.0',
            metrics_port: int = 8887,
            **kwargs
    ):
        super().__init__(name=self.__class__.__name__, timeout=0.1, **kwargs)
        logging.getLogger().setLevel(logging.INFO)
        self._pipeline = pipeline
        self._build_pipeline()
        self.start()

        if verbose_metrics:
            self._templates = Jinja2Templates(directory=join(dirname(abspath(__file__)), 'templates'))
            app = FastAPI()
            app.mount('/static', StaticFiles(directory=join(dirname(abspath(__file__)), 'static'), html=True), 'static')
            router = APIRouter()
            router.add_api_route('/metrics_html', self._metrics_html, methods=['GET'], response_class=HTMLResponse)
            router.add_api_route('/metrics_json', self._metrics_json, methods=['GET'])
            app.include_router(router)
            Thread(target=uvicorn.run, kwargs=dict(app=app, host=metrics_host, port=metrics_port)).start()

    def _metrics_html(self, request: Request):
        return self._templates.TemplateResponse('home.html', dict(request=request))

    def _metrics_json(self):
        metrics = []

        for node in self._pipeline:
            metrics.append(dict(name=node.name,
                                input=node.metrics['input'],
                                output=node.metrics['output'],
                                duration_mean_ms=f"{mean(node.metrics['duration']) * 1000:.2f}",
                                duration_std_ms=f"{pstdev(node.metrics['duration']) * 1000:.2f}"))

        return dict(current_time=datetime.now().strftime('%Y-%m-%d %H:%M:%S'), metrics=metrics)

    def _build_pipeline(self):
        queues = {}

        for node in self._pipeline:
            for queue in node.input + node.output:
                if queue in queues:
                    continue

                queues[queue] = QueueWrapper(name=queue, type=node.type, maxsize=16)

        for node in self._pipeline:
            for queue in node.input:
                node.input_queue.append(queues[queue])

            for queue in node.output:
                node.output_queue.append(queues[queue])

            node.start()

    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        if all(node.is_finish for node in self._pipeline):
            for node in self._pipeline:
                node.release()

            logging.info(f'{self.name}: All nodes finished successfully')

            return Batch(status=BatchStatus.END)

        elif any(node.is_terminate for node in self._pipeline):
            logging.info(f'{self.name}: At least one of the nodes has terminated')

            for node in self._pipeline:
                node.release()
                node.drain()
                logging.info(f'{node.name}: Draining')

            return Batch(status=BatchStatus.END)



from typing import Tuple
import cv2
import numpy as np


def resize(
        image: np.ndarray,
        size: Tuple[int, int],
        pad_value: int = 0
) -> np.ndarray:
    h, w, c = image.shape

    if (h, w) == (size[0], size[1]):
        return image

    ratio = min(size[0] / h, size[1] / w)
    h_new, w_new = round(h * ratio), round(w * ratio)

    image_resized = cv2.resize(image, (0, 0), fx=ratio, fy=ratio, interpolation=cv2.INTER_AREA)
    image_padded = np.full((size[0], size[1], c), fill_value=pad_value, dtype=image.dtype)

    h_pos, w_pos = (size[0] - h_new) // 2, (size[1] - w_new) // 2
    image_padded[h_pos: h_pos + h_new, w_pos: w_pos + w_new] = image_resized

    return image_padded



from typing import Optional, List, Dict
from pipez.core.enums import BatchStatus


class Batch(object):
    def __init__(
            self,
            data: Optional[List[Dict]] = None,
            meta: Optional[Dict] = None,
            status: BatchStatus = BatchStatus.OK,
            error: Optional[str] = None
    ):
        self._data = data if data else []
        self._meta = meta.copy() if meta else {}
        self._status = status
        self._error = error

    @property
    def data(self) -> List[Dict]:
        return self._data

    @property
    def meta(self) -> Dict:
        return self._meta

    @property
    def is_ok(self):
        return self._status == BatchStatus.OK

    @property
    def is_end(self):
        return self._status == BatchStatus.END

    @property
    def is_error(self):
        return self._status == BatchStatus.ERROR

    @property
    def error(self) -> str:
        return self._error

    def __getitem__(self, item) -> Dict:
        return self._data[item]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def append(self, data: Dict):
        self._data.append(data)



class Registry(object):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
            cls._instance._registry = {}

        return cls._instance

    def __getitem__(self, item):
        if item not in self._registry:
            raise KeyError(f'Class «{item}» not found in registry, it can be added using @Registry.add')

        return self._registry[item]

    @staticmethod
    def add(cls):
        Registry()._registry[cls.__name__] = cls
        return cls



from abc import ABC, abstractmethod
from typing import Optional
from fastapi import FastAPI, APIRouter
from fastapi.openapi.docs import get_swagger_ui_html, get_swagger_ui_oauth2_redirect_html
from fastapi.staticfiles import StaticFiles
from os.path import join, dirname, abspath
import uvicorn

from pipez.core.batch import Batch
from pipez.core.node import Node


class NodeFastAPI(Node, ABC):
    def __init__(
            self,
            host: str = '0.0.0.0',
            port: int = 8888,
            **kwargs
    ):
        super().__init__(**kwargs)
        self._host = host
        self._port = port
        self._app = FastAPI(docs_url=None)
        self._router = APIRouter()

    def _mount_localhost_ui(self):
        self._app.mount('/static', StaticFiles(directory=join(dirname(abspath(__file__)), 'SwaggerUI')), 'static')

        @self._app.get('/docs', include_in_schema=False)
        async def custom_swagger_ui_html():
            return get_swagger_ui_html(openapi_url=self._app.openapi_url,
                                       title=f'{self._app.title} - Swagger UI',
                                       swagger_js_url='/static/swagger-ui-bundle.js',
                                       swagger_css_url='/static/swagger-ui.css',
                                       oauth2_redirect_url=self._app.swagger_ui_oauth2_redirect_url)

        @self._app.get(self._app.swagger_ui_oauth2_redirect_url, include_in_schema=False)
        async def swagger_ui_redirect():
            return get_swagger_ui_oauth2_redirect_html()

    @abstractmethod
    def add_api_routes(self):
        pass

    def processing(self, input: Optional[Batch]) -> Optional[Batch]:
        self._mount_localhost_ui()
        self.add_api_routes()
        self._app.include_router(self._router)
        uvicorn.run(self._app, host=self._host, port=self._port)

        return None



from typing import List, Dict

from pipez.core.node import Node
from pipez.core.registry import Registry
from pipez.core.enums import NodeType


def json2nodes(
        pipeline: List[Dict]
) -> List[Node]:
    registry = Registry()
    nodes = []

    for node in pipeline:
        if not isinstance(node, Dict):
            raise TypeError('JSON format must be class «dict»')

        cls = registry[node.pop('cls')]
        node['type'] = NodeType.from_string(node.get('type', 'thread'))
        nodes.append(cls(**node))

    return nodes



from abc import ABC, abstractmethod
from typing import Optional, List, Dict, Union
from threading import Thread
from multiprocessing import Process
from collections import deque
import logging
import time

from pipez.core.batch import Batch
from pipez.core.enums import NodeType, NodeStatus, BatchStatus
from pipez.core.memory import Memory
from pipez.core.queue_wrapper import QueueWrapper


class Node(ABC):
    def __init__(
            self,
            name: str,
            type: NodeType = NodeType.THREAD,
            input: Optional[Union[str, List[str]]] = None,
            output: Optional[Union[str, List[str]]] = None,
            timeout: float = 0.0
    ):
        self._name = name
        self._type = type

        if self._type == NodeType.THREAD:
            self._worker = Thread(target=self._run, name=self._name)

        elif self._type == NodeType.PROCESS:
            self._worker = Process(target=self._run, name=self._name)

        self._input = [input] if isinstance(input, str) else input if input else []
        self._output = [output] if isinstance(output, str) else output if output else []
        self._input_queue = []
        self._output_queue = []

        self._memory = Memory()
        self._metrics = dict(duration=deque([0], maxlen=100), input=0, output=0)
        self._timeout = timeout
        self._status = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> NodeType:
        return self._type

    @property
    def input(self) -> List[str]:
        return self._input

    @property
    def output(self) -> List[str]:
        return self._output

    @property
    def input_queue(self) -> List[QueueWrapper]:
        return self._input_queue

    @property
    def output_queue(self) -> List[QueueWrapper]:
        return self._output_queue

    @property
    def memory(self) -> Memory:
        return self._memory

    @property
    def metrics(self) -> Dict:
        return self._metrics

    @property
    def is_alive(self) -> bool:
        return self._status == NodeStatus.ALIVE

    @property
    def is_finish(self) -> bool:
        return self._status == NodeStatus.FINISH

    @property
    def is_terminate(self) -> bool:
        return self._status == NodeStatus.TERMINATE

    def start(self):
        self._status = NodeStatus.ALIVE
        self._worker.start()

    def _get(self) -> Optional[Batch]:
        if not self._input_queue:
            return None

        if len(self._input_queue) == 1:
            return self._input_queue[0].get()

        batches = [queue.get() for queue in self._input_queue]

        if len(set(len(batch) for batch in batches)) != 1:
            return Batch(status=BatchStatus.ERROR, error='Length batches cannot be different')

        if all(batch.is_ok for batch in batches):
            status = BatchStatus.OK
        elif all(batch.is_end for batch in batches):
            status = BatchStatus.END
        else:
            return Batch(status=BatchStatus.ERROR, error='Batches cannot have different status')

        meta = {}

        for batch in batches:
            meta.update(batch.meta)

        data = [
            {
                queue.name: batch[idx]
                for queue, batch in zip(self._input_queue, batches)
            }
            for idx in range(len(batches[0]))
        ]

        return Batch(data=data, meta=meta, status=status)

    def _step(self, input: Optional[Batch]) -> Optional[Batch]:
        try:
            monotonic = time.monotonic()
            output = self.processing(input)
            self._metrics['duration'].append(time.monotonic() - monotonic)
            self._metrics['input'] += len(input) if isinstance(input, Batch) else 0
            self._metrics['output'] += len(output) if isinstance(output, Batch) else 0
        except Exception as e:
            output = Batch(status=BatchStatus.ERROR, error=f'During processing raise exception {e.__class__} {e}')

        return output

    def _put(self, output: Batch):
        if not self._output_queue:
            return

        for queue in self._output_queue:
            queue.put(output)

    def _run(self):
        while True:
            time.sleep(self._timeout)

            if not self.is_alive:
                break

            input = self._get()

            if isinstance(input, Batch):
                if input.is_end:
                    self._put(input)
                    self._status = NodeStatus.FINISH
                    break
                elif input.is_error:
                    logging.error(f'{self.name}: {input.error}')
                    self._status = NodeStatus.TERMINATE
                    break

            output = self._step(input)

            if isinstance(output, Batch):
                if output.is_ok:
                    self._put(output)
                elif output.is_end:
                    self._put(output)
                    self._status = NodeStatus.FINISH
                    break
                elif output.is_error:
                    logging.error(f'{self.name}: {output.error}')
                    self._status = NodeStatus.TERMINATE
                    break

    def release(self):
        pass

    def drain(self):
        self._status = NodeStatus.TERMINATE

        for queue in self._input_queue + self._output_queue:
            while not queue.empty():
                queue.get()

    @abstractmethod
    def processing(self, data: Optional[Batch]) -> Optional[Batch]:
        pass



from setuptools import setup, find_packages
from os.path import join, dirname
from pipez import __version__


setup(
    name='pipez',
    version=__version__,
    author="Alexander Timofeev",
    author_email="tam2511@mail.ru",
    python_requires=">=3.8",
    packages=find_packages(),
    long_description=open(join(dirname(__file__), 'README.md')).read(),
    long_description_content_type="text/markdown",
    url="https://github.com/tam2511/pipez",
    install_requires=[
        'numpy',
        'opencv-python',
        'fastapi',
        'uvicorn',
        'Jinja2'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
    include_package_data=True
)
